#pragma once

#include <stddef.h>

#include "nvim/vterm/vterm_defs.h"
#include "nvim/vterm/vterm_keycodes_defs.h"

#ifdef INCLUDE_GENERATED_DECLARATIONS
# include "vterm/keyboard.h.generated.h"
#endif
