#pragma once

#include "nvim/highlight_defs.h"  // IWYU pragma: keep
#include "nvim/tui/tui_defs.h"  // IWYU pragma: keep
#include "nvim/ui_defs.h"  // IWYU pragma: keep

#ifdef INCLUDE_GENERATED_DECLARATIONS
# include "tui/tui.h.generated.h"
#endif
